#include "signals.h"
#include "classes.h"
#include "utils.h"

JSObject *signal_arg_map;

typedef struct {
	char *signal;
	char *args[7];
	int dynamic;
} PERL_SIGNAL_ARGS_REC;

#include <perl/perl-signals-list.h>

jsval
conv_irssi_arg(JSContext *cx, const char *arg_type, const void *arg, JSBool *ref);

jsval
conv_irssi_arg(JSContext *cx, const char *at, const void *arg, JSBool *ref) {
  if (ref) *ref = JS_FALSE;
  switch (at[0]) {
  case 'c': /* char* */
    return safe_make_js_string(cx, arg);
  case 's':
    switch (at[1]) {
    case 't': /* string */
      return safe_make_js_string(cx, arg);
      
    case 'i': /* siobject */
      return OBJECT_TO_JSVAL(gjs_irssi_c_bless_object(cx, at, arg));
    }
    break;
  case 'i':
    switch (at[1]) {
    case 'o': /* iobject */;
      return OBJECT_TO_JSVAL(gjs_irssi_c_bless_object(cx, at, arg));

    case 'n':
      if (strlen(at) < 4) /* int */
	return INT_TO_JSVAL((intptr_t)arg);
      else { /* intptr */
	JSObject *p;
	if (!(p = JS_NewObject(cx, &gjs_value_reference_class, gjs_value_reference_proto, NULL)))
	  break;
	if (!JS_DefineProperty(cx, p, "value", INT_TO_JSVAL(*(int *) arg), NULL, NULL, JSPROP_ENUMERATE))
	  break;
	if (ref) *ref = JS_TRUE;
	return OBJECT_TO_JSVAL(p);
      }
    }
    break;
  case '\0':
    break;
  case 'g':
    switch (at[1]) {
    case 'l': /* glistptr_ */
      if (strlen(at) <= 9)
	break;
      {
	GList *tmp;
	JSObject *ret, *p;
	uint32_t  i = 0;
	if (!(ret = JS_NewArrayObject(cx, 0, NULL)))
	  break;
	
	for (tmp = *(GList**)arg; tmp != NULL; tmp = tmp->next) {
	  jsval r;
	  r = conv_irssi_arg(cx, at + 9, tmp->data, NULL);
	  JS_SetElement(cx, ret, i++, &r);
	}
	if (!(p = JS_NewObject(cx, &gjs_value_reference_class, gjs_value_reference_proto, NULL)))
	  break;
	if (!JS_DefineProperty(cx, p, "value", OBJECT_TO_JSVAL(ret), NULL, NULL, JSPROP_ENUMERATE))
	  break;
	if (ref) *ref = JS_TRUE;
	return OBJECT_TO_JSVAL(p);
      }
    case 's': /* gslist_ */
      if (strlen(at) <= 7)
	break;
      {
	GSList *tmp;
	JSObject *ret;
	uint32_t  i = 0;
	ret = JS_NewArrayObject(cx, 0, NULL);

	for (tmp = (GSList*)arg; tmp != NULL; tmp = tmp->next) {
	  jsval r;
	  r = conv_irssi_arg(cx, at + 7, tmp->data, NULL);
	  JS_SetElement(cx, ret, i++, &r);
	}
	return OBJECT_TO_JSVAL(ret);
      }
    }
    break;
  case 'u': /* ulongptr */
    return JS_NumberValue(*(unsigned long *) arg);
  default:
    return OBJECT_TO_JSVAL(gjs_irssi_c_bless_object(cx, at, arg));
  }
  return JSVAL_VOID;
}

void conv_irssi_arg_back(JSContext *cx, const char *at, const void *arg, const jsval *vp) {
  switch (at[0]) {
  case 'i': /* intptr */
    {
      jsval res;
      int32_t new_value;
      if (JSVAL_IS_PRIMITIVE(*vp))
	return;
      if (!JS_GetProperty(cx, JSVAL_TO_OBJECT(*vp), "value", &res))
	return;
      if (!JS_ValueToInt32(cx, res, &new_value))
	return;
      (*(int *) arg) = new_value;
      return;
    }
  case 'g': /* glistptr_ */
    if (strlen(at) <= 9)
      return;
    {
      jsval res;
      GList **ret = (GList **)arg;
      GList *out = NULL;
      uint32_t  i, len;
      JSBool string_value = JS_FALSE, int_value = JS_FALSE, unsigned_value = JS_FALSE;
      
      if (JSVAL_IS_PRIMITIVE(*vp))
	return;
      if (!JS_GetProperty(cx, JSVAL_TO_OBJECT(*vp), "value", &res))
	return;
      if (JSVAL_IS_PRIMITIVE(res) || !(JS_GetArrayLength(cx, JSVAL_TO_OBJECT(res), &len)))
	return;

      switch (at[9]) {
      case 'c':
      case 's':
	string_value = JS_TRUE;
	break;
      case 'i':
	int_value = JS_TRUE;
	break;
      case 'u':
	unsigned_value = JS_TRUE;
	break;
      default:
	return;
      }
      for (i = 0; i < len; ++i) {
	jsval val;
	void *gval;
	if (!JS_GetElement(cx, JSVAL_TO_OBJECT(res), i, &val))
	  goto prop_err;

	if (string_value) {
	  gval = decode_js_string(cx, JS_ValueToString(cx, val));
	}
	else if (int_value) {
	  int ival;
	  if (!JS_ValueToECMAInt32(cx, val, &ival))
	    continue;
	  gval = GINT_TO_POINTER(ival);
	}
	else if (unsigned_value) {
	  unsigned uval;
	  if (!JS_ValueToECMAUint32(cx, val, &uval))
	    continue;
	  gval = GUINT_TO_POINTER(uval);
	}
	out = g_list_prepend(out, gval);
      }
      out = g_list_reverse(out);

      if (string_value)
	g_list_free_full(*ret, g_free);
      else
	g_list_free(*ret);
      
      *ret = out;

      return;
    prop_err:
      if (string_value)
	g_list_free_full(out, g_free);
      else
	g_list_free(out);
      return;
    }
  }
}

void gjs_irssi_sig_proxy (const void *p1, const void *p2,
                          const void *p3, const void *p4,
                          const void *p5, const void *p6) {
  jsval sobj, arg_info;
  jsval argv[6];
  JSBool reference_type[6];
  uint32_t argc = 6;
  const void *p[6];
  JSObject *info = signal_get_user_data();
  nCTX;
  nSCX;
  uSCX;
  unsigned int i;
  
  for (i = 0; i < 6; ++i) { argv[i] = JSVAL_VOID; }
  p[0] = p1, p[1] = p2, p[2] = p3, p[3] = p4, p[4] = p5, p[5] = p6;

  JS_BeginRequest(cx);
  {
    jsval is_running;
    if (!JS_GetProperty(cx, info, "running", &is_running))
      goto end;
    if (JSVAL_IS_BOOLEAN(is_running) && JSVAL_TO_BOOLEAN(is_running))
      goto end;
  }
  
  JS_DefineProperty(cx, info, "running", JSVAL_TRUE, NULL, NULL, FIXED_PROP_FLAGS);
  get_pSCX(info, "script_name");
  get_sObj(sobj);

  arg_info = JSVAL_VOID;
  {
    char *partial_signal, *ps_pos = NULL;
    partial_signal = g_strdup(signal_get_emitted());
    while (*partial_signal) {
      if (!JS_GetProperty(cx, signal_arg_map, partial_signal, &arg_info))
	goto sObj_cleanup;

      if (!JSVAL_IS_VOID(arg_info))
	break;

      if (ps_pos)
	ps_pos[0] = '\0';

      ps_pos = strrchr(partial_signal, ' ');

      if (!ps_pos)
	break;

      ps_pos[1] = '\0';
    }
    g_free(partial_signal);
  }

  if (JSVAL_IS_VOID(arg_info)) {
    g_message("\"%s\" is not a registered signal", signal_get_emitted());
    goto sObj_cleanup;
  }
  
  for (i = 0; i < 6; ++i) {
    jsval parm_type;
    char *st;

    if (!JS_GetElement(cx, JSVAL_TO_OBJECT(arg_info), i, &parm_type))
      goto sObj_cleanup;

    if (JSVAL_IS_VOID(parm_type))
      break;

    if (!JSVAL_IS_STRING(parm_type))
      continue;
      
    st = decode_js_string(cx, JSVAL_TO_STRING(parm_type));
    argv[i] = conv_irssi_arg(cx, st, p[i], &reference_type[i]);
    g_free(st);
  }
  argc = i;

  StocSCX;

  {
    jsval callback, rval;
    if (!JS_GetProperty(cx, info, "callback", &callback))
      goto prop_err;
    if (JSVAL_IS_STRING(callback)) {
      char *st;
      JS_CallFunctionName(cx, JSVAL_TO_OBJECT(sobj),
			  st = JS_EncodeString(cx, JSVAL_TO_STRING(callback)),
			  argc, argv, &rval);
      JS_free(cx, st);
    }
    else if (!JSVAL_IS_PRIMITIVE(callback)) {
      JS_CallFunctionValue(cx, JSVAL_TO_OBJECT(sobj), callback, argc, argv, &rval);
    }
  }

  for (i = 0; i < argc; ++i) {
    if (reference_type[i]) {
      jsval parm_type;
      char *st;

      if (!JS_GetElement(cx, JSVAL_TO_OBJECT(arg_info), i, &parm_type))
	goto prop_err;

      if (JSVAL_IS_VOID(parm_type))
	break;

      if (!JSVAL_IS_STRING(parm_type))
	continue;
      
      st = decode_js_string(cx, JSVAL_TO_STRING(parm_type));
      conv_irssi_arg_back(cx, st, p[i], &argv[i]);
      g_free(st);
    }
  }
 prop_err:
  RstSCX;
 sObj_cleanup:
 SCX_cleanup:
  JS_DefineProperty(cx, info, "running", JSVAL_VOID, NULL, NULL, 0);
  JS_DeleteProperty(cx, info, "running");
 end:
  if (JS_IsExceptionPending(cx))
    JS_ClearPendingException(cx);
  JS_EndRequest(cx);
  gjs_maybe_gc(cx);
}

SCRIPT_FUN(irssi_signal_add) {
    nRV;
    nArgv;
    nSCX;
    jsval temp_val, callback;
    JSString *str;
    char *st, *st2, *st3;
    JSObject *rec, *temp_obj;
    int32_t prio = SIGNAL_PRIORITY_DEFAULT;

    JS_BeginRequest(cx);

    if (!JS_ConvertArguments(cx, argc, argv, "Sv/j", &str, &callback, &prio)) {
        JS_ReportError(cx, "argument mismatch");
        goto out;
    }

    st = JS_EncodeString(cx, str);

    get_cSCX;
    
    rec = JS_NewObject(cx, NULL, NULL, NULL);
    JS_DefineProperty(cx, rec, "signal", STRING_TO_JSVAL(str), NULL, NULL, FIXED_PROP_FLAGS);
    JS_DefineProperty(cx, rec, "script_name", sCX, NULL, NULL, FIXED_PROP_FLAGS);
    JS_DefineProperty(cx, rec, "callback", callback, NULL, NULL, FIXED_PROP_FLAGS);
    JS_DefineProperty(cx, rec, "priority", INT_TO_JSVAL(prio), NULL, NULL, FIXED_PROP_FLAGS);

    if (!JS_GetProperty(cx, gjs_irssi_script_loc, "Signals", &temp_val))
        goto cleanup;
    if (JSVAL_IS_PRIMITIVE(temp_val))
        temp_obj = JS_DefineObject(cx, gjs_irssi_script_loc, "Signals", NULL, NULL, FIXED_PROP_FLAGS);
    else
        temp_obj = JSVAL_TO_OBJECT(temp_val);
    if (!temp_obj) {
        g_message("Could not hold Signals data");
        goto cleanup;
    }
    if (!JS_GetProperty(cx, temp_obj, st2 = JS_EncodeString(cx, JSVAL_TO_STRING(sCX)), &temp_val))
        goto cleanup2;
    if (JSVAL_IS_PRIMITIVE(temp_val))
        temp_obj = JS_DefineObject(cx, temp_obj, st2, NULL, NULL, FIXED_PROP_FLAGS);
    else
        temp_obj = JSVAL_TO_OBJECT(temp_val);
    if (!temp_obj)
        goto cleanup2;
    if (!JS_GetProperty(cx, temp_obj, st, &temp_val))
        goto cleanup2;
    if (JSVAL_IS_PRIMITIVE(temp_val)) {
        temp_obj = JS_DefineObject(cx, temp_obj, st, NULL, NULL, FIXED_PROP_FLAGS);
        JS_DefineProperty(cx, temp_obj, "length", JSVAL_ZERO, NULL, NULL, HID_FIX_PROP_FLAGS);
    }
    else {
        temp_obj = JSVAL_TO_OBJECT(temp_val);
    }
    if (!temp_obj)
        goto cleanup2;

    st3 = g_strdup_printf("%p", (void*)rec);
    JS_DefineProperty(cx, temp_obj, st3, OBJECT_TO_JSVAL(rec), NULL, NULL, FIXED_PROP_FLAGS);
    g_free(st3);
    if (!JS_GetProperty(cx, temp_obj, "length", &temp_val))
        goto cleanup2;
    if (!JSVAL_IS_INT(temp_val))
        goto cleanup2;
    JS_DefineProperty(cx, temp_obj, "length", INT_TO_JSVAL( JSVAL_TO_INT(temp_val) + 1 ), NULL, NULL, HID_FIX_PROP_FLAGS);
    rv = JS_TRUE;
    signal_add_full(MODULE_NAME, prio, st,
                    (SIGNAL_FUNC)gjs_irssi_sig_proxy, rec);
    JS_SET_RVAL(cx, vp, OBJECT_TO_JSVAL(rec));    
 cleanup2:
    JS_free(cx, st2);
 cleanup:
    JS_free(cx, st);
 SCX_cleanup:
 out:
    JS_EndRequest(cx);
    return rv;
}

JSBool
script_signal_remove_all(JSContext *cx, const char *data) {
  jsval temp_val;
  char *st2;
  JSObject *sign_it, *sigp_it, *rec, *temp_obj, *script_so, *script_sso;
  jsid sign, sigp;
    
  if (!JS_GetProperty(cx, gjs_irssi_script_loc, "Signals", &temp_val))
    goto out;
  if (JSVAL_IS_PRIMITIVE(temp_val))
    goto out;
  temp_obj = JSVAL_TO_OBJECT(temp_val);
  if (!JS_GetProperty(cx, temp_obj, data, &temp_val))
    goto out;
  if (JSVAL_IS_PRIMITIVE(temp_val))
    goto out;

  script_so = JSVAL_TO_OBJECT(temp_val);
  sign_it = JS_NewPropertyIterator(cx, script_so);
  if (!sign_it)
    goto out;

  while (JS_NextProperty(cx, sign_it, &sign) && sign != JSID_VOID) {
    if (!JS_GetPropertyById(cx, script_so, sign, &temp_val))
      continue;
    if (JSVAL_IS_PRIMITIVE(temp_val))
      continue;
    script_sso = JSVAL_TO_OBJECT(temp_val);
    sigp_it = JS_NewPropertyIterator(cx, script_sso);
    if (!sigp_it)
      continue;

    while (JS_NextProperty(cx, sigp_it, &sigp) && sigp != JSID_VOID) {
      if (!JS_GetPropertyById(cx, script_sso, sigp, &temp_val))
	continue;
      if (JSVAL_IS_PRIMITIVE(temp_val))
	continue;
        
      rec = JSVAL_TO_OBJECT(temp_val);
      if (!JS_GetProperty(cx, rec, "signal", &temp_val))
	continue;
      if (!JSVAL_IS_STRING(temp_val))
	continue;
      st2 = JS_EncodeString(cx, JSVAL_TO_STRING(temp_val));
      signal_remove_full(st2, (SIGNAL_FUNC)gjs_irssi_sig_proxy, rec);

      JS_free(cx, st2);
    }
  }
  JS_DefineProperty(cx, temp_obj, data, JSVAL_VOID, NULL, NULL, 0);
  JS_DeleteProperty(cx, temp_obj, data);
    
 out:
  return JS_TRUE;
}

SCRIPT_FUN(irssi_signal_remove) {
  nArgv;
  nSCX;
  nRV;

  jsval temp_val, signal, func;
  JSObject *temp_obj, *script_signal_obj, *specific_signal_obj, *rec = NULL;
  char *st2, *signal_pn = NULL;
  JSBool succ;

  JS_BeginRequest(cx);
    
  get_cSCX;
    
  signal = JSVAL_VOID;
  if (!JSVAL_IS_PRIMITIVE(argv[0])) {
    rec = JSVAL_TO_OBJECT(argv[0]);
    if (!JS_GetProperty(cx, rec, "signal", &signal))
      goto out;
  }
  else if (JSVAL_IS_STRING(argv[0]) && argc > 1) {
    signal = argv[0];
    func = argv[1];
  }
  if (JSVAL_IS_VOID(signal)) {
    JS_ReportError(cx, "Not a signal object");
    goto out;
  }

  if (!JS_GetProperty(cx, gjs_irssi_script_loc, "Signals", &temp_val))
    goto out;
  if (JSVAL_IS_PRIMITIVE(temp_val))
    goto out;

  temp_obj = JSVAL_TO_OBJECT(temp_val);
  succ = JS_GetProperty(cx, temp_obj, st2 = JS_EncodeString(cx, JSVAL_TO_STRING(sCX)), &temp_val);
  JS_free(cx, st2);
  if (!succ)
    goto out;
  if (JSVAL_IS_PRIMITIVE(temp_val))
    goto out;
  script_signal_obj = JSVAL_TO_OBJECT(temp_val);
  succ = JS_GetProperty(cx, script_signal_obj, st2 = JS_EncodeString(cx, JSVAL_TO_STRING(signal)), &temp_val);
  JS_free(cx, st2);
  if (!succ)
    goto out;
  if (JSVAL_IS_PRIMITIVE(temp_val)) {
    JS_ReportWarning(cx, "%s", "Signal not found");
    goto out;
  }
  specific_signal_obj = JSVAL_TO_OBJECT(temp_val);

  if (rec) {
    succ = JS_GetProperty(cx, specific_signal_obj, signal_pn = g_strdup_printf("%p", (void*)rec), &temp_val);
    if (!succ)
      goto out;
    if (!JSVAL_IS_PRIMITIVE(temp_val))
      rec = JSVAL_TO_OBJECT(temp_val);
    else
      rec = NULL;
  }
  else {
    jsid id;
    jsval sig;
    int32_t res;
    JSObject *sig_it = JS_NewPropertyIterator(cx, specific_signal_obj);
    if (!sig_it)
      goto out;
    while (JS_NextProperty(cx, sig_it, &id) && id != JSID_VOID) {
      if (!JS_GetPropertyById(cx, specific_signal_obj, id, &sig))
	continue;
      if (JSVAL_IS_PRIMITIVE(sig))
	continue;
      if (!JS_GetProperty(cx, JSVAL_TO_OBJECT(sig), "callback", &temp_val))
	continue;
      if (JSVAL_IS_STRING(temp_val) && JSVAL_IS_STRING(func)) {
	if (!JS_CompareStrings(cx, JSVAL_TO_STRING(temp_val), JSVAL_TO_STRING(func), &res))
	  continue;
	if (res == 0) {
	  rec = JSVAL_TO_OBJECT(sig);
	  signal_pn = g_strdup_printf("%p", (void*)rec);
	  break;
	}
      }
      else if (!JSVAL_IS_PRIMITIVE(temp_val) && !JSVAL_IS_PRIMITIVE(func)
	       && JSVAL_TO_OBJECT(temp_val) == JSVAL_TO_OBJECT(func)) {
	rec = JSVAL_TO_OBJECT(sig);
	signal_pn = g_strdup_printf("%p", (void*)rec);
	break;
      }
    }
  }
        
  if (!rec) {
    JS_ReportWarning(cx, "%s", "Signal not found");
    goto out;
  }
  st2 = JS_EncodeString(cx, JSVAL_TO_STRING(signal));
  signal_remove_full(st2, (SIGNAL_FUNC)gjs_irssi_sig_proxy, rec);
  if (signal_pn) {
    JS_DefineProperty(cx, specific_signal_obj, signal_pn, JSVAL_VOID, NULL, NULL, 0);
    JS_DeleteProperty(cx, specific_signal_obj, signal_pn);
  }

  if (!JS_GetProperty(cx, specific_signal_obj, "length", &temp_val))
    goto out2;
  if (!JSVAL_IS_INT(temp_val))
    goto out2;
  if (JSVAL_TO_INT(temp_val) == 1) {
    JS_DefineProperty(cx, script_signal_obj, st2, JSVAL_VOID, NULL, NULL, 0);
    JS_DeleteProperty(cx, script_signal_obj, st2);
  }
  else
    JS_DefineProperty(cx, specific_signal_obj, "length",
		      INT_TO_JSVAL( JSVAL_TO_INT(temp_val) - 1 ),
		      NULL, NULL, HID_FIX_PROP_FLAGS);
  rv = JS_TRUE;
  JS_SET_RVAL(cx, vp, JSVAL_VOID);    
 out2:
  JS_free(cx, st2);
 out:
 SCX_cleanup:
  if (signal_pn)
    g_free(signal_pn);
  JS_EndRequest(cx);
  return rv;
}

JSBool
init_signals(JSContext *cx, JSObject *global) {
  nRV;
  PERL_SIGNAL_ARGS_REC *sr;
  jsval v;
  JSObject *o;
  
  JS_BeginRequest(cx);

  if (!JS_GetProperty(cx, global, "Irssi", &v))
    goto out;

  if (!JS_GetProperty(cx, JSVAL_TO_OBJECT(v), "prototype", &v))
    goto out;
  
  if (!(o = JS_DefineObject(cx, JSVAL_TO_OBJECT(v), "SignalMap", NULL, NULL, FIXED_PROP_FLAGS)))
    goto out;
  
  for (sr = perl_signal_args; sr->signal; ++sr) {
    char **args;
    JSObject *ar;
    unsigned i = 0;
    ar = JS_NewArrayObject(cx, 0, NULL);
    for (args = sr->args; *args; ++args) {
      jsval r = safe_make_js_string(cx, *args);
      JS_SetElement(cx, ar, i++, &r);
    }
    JS_FreezeObject(cx, ar);
    JS_DefineProperty(cx, o, sr->signal, OBJECT_TO_JSVAL(ar), NULL, NULL, FIXED_PROP_FLAGS);
  }
  signal_arg_map = o;
  rv = JS_TRUE;
 out:
  JS_EndRequest(cx);
  return rv;
}
